/* $Id$ */
/***********************************************************************
 Filename:	offline_cmd_hist.c
 Purpose:	OFFLINE merge the Command History Report with existing
                cmd hist reports for the same operational day. 
                NOTE: Merge in time-order with no duplicates!		       

 Contains:	offline_format_cmd_hist
		offline_get_op_day
		offline_merge_cmdhist
                offline_remove_duplicates
                offline_sort_cmdhist
                offline_sort_bgq
***********************************************************************/
/* $Log$ */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/wait.h>
#include "mission.h"
#include "iws.h"
#include "delayed_events.h"

/* #define DEBUG 1 */

/* externals */
extern char *cmd_hist_dir;
extern char *temp_files_dir;

/* function prototypes */
int append_files(char *file1, char *file2, char *file3);
int offline_format_cmd_hist(char *final_cmd_hist, char *final_cmd, char *final_bck);
int offline_get_op_day(char *filename, char *filedate);
int offline_merge_cmdhist(char *filedate, char *scihist2, char *scibck2, 
                  char *final_cmd_hist);
int offline_remove_duplicates(char *infile, char *outfile);
int offline_sort_cmdhist(char *infile, char *outfile);
int offline_sort_bgq(char *infile, char *outfile);
int offline_split_cmdhist(char *cmdhist, char *filedate, char *scihist2,char *scibck2);


/*********************************************************************
 FUNCTION NAME:         offline_split_cmdhist
 INPUT  PARAMETERS:     char *cmdhist,char *filedate,
                        char *scihist2,char *scibck2
 OUTPUT PARAMETERS:     char *filedate, char *scihist2, char *scibck2
 RETURN VALUE:          int rc (OK,ERR)
 REFERENCED DATA FILES: cmdhist (POCC command history file )
                        scihist2 (cmd hist msgs from cmdhist)
                        scibck2  (background queue  msgs from cmdhist )
 PURPOSE:               Split the cmdhist msgs into two files 
                        (cmd hist and bgq) for
                        merging into existing Cmd History Rpt.
  *********************************************************************/
int offline_split_cmdhist(char *cmdhist, char *filedate, char *scihist2, char *scibck2)
{
   FILE *infile;
   FILE *cmdfile;
   FILE *bckqfile;
   int rc;
   char *ptr;
   char string_buffer[MAXBUFFER] = {'\0'};
   char errmsg[MAXBUFFER] = {'\0'};
   char buf[MAXBUFFER];
   char cmdhist_filename[MAXBUFFER];
   bool found_data =    FALSE;

   /* open the POCC command history report file */
   if ((infile = fopen(cmdhist, "r")) == NULL)
   {
      printf( "Cannot Open File %s\n",cmdhist);
      return(ERR);
   }

   /* convert POCC naming convention (dddhhmmss) to
      EOF (yyyymmdd) format */
   if (ptr = strrchr(cmdhist,'/'))
   {
      strcpy(cmdhist_filename,++ptr);
   }
   else
   {
       strcpy(cmdhist_filename,cmdhist);
   }
   offline_get_op_day(cmdhist_filename, filedate);

   /* get EOF cmd hist and bckgrnd Q report names */
   sprintf(scihist2,"%s/%s_V02.CHR",temp_files_dir,filedate);
   sprintf(scibck2, "%s/%s_V02.BGQ",temp_files_dir,filedate);

   /* open the EOF command history report */
   if ((cmdfile = fopen(scihist2, "w")) == NULL)
   {
      printf("Cannot open %s\n",scihist2);
      return(ERR);
   }

   /* open the EOF background queue report */
   if ((bckqfile = fopen(scibck2, "w")) == NULL)
   {
      printf("Cannot open %s\n",scibck2);
      return(ERR);
   }

   /* split the POCC cmd hist report into two files
            yyyymmdd_V02.CHR and yyyymmdd_V02.BGQ */
   while (!feof(infile))
   {
      if (fgets(string_buffer,MAXBUFFER-1,infile) == NULL)
      {
         break;
      }

      if (strstr(string_buffer, "BACKGROUND QUEUE REPORT"))
      {  /* found end of cmd hist report */
         /* now append bck Q rpt file */

         while (!feof(infile))
         {
            if (fgets(string_buffer,MAXBUFFER-1,infile) == NULL)
            {
               break;
            }
            if (found_data)
            {  /* ignore format lines at top/bottom */
               if (!EQU(string_buffer,"\n") &&
                  (strstr(string_buffer,"END OF BACKGROUND QUEUE REPORT") == NULL) &&
                  (strchr(string_buffer,'*') == NULL) &&
                  (strstr(string_buffer,"CMS Background Queue:") == NULL) &&
                  (strstr(string_buffer,"Current Mode Parameter Values") == NULL) &&
                  (strstr(string_buffer,"Throughput Mode:") == NULL && strstr(string_buffer,"RCR Processing:") == NULL) &&
                  (strstr(string_buffer,"Mission:")  == NULL && strstr(string_buffer,"Local GMT:") == NULL ) &&
                  (strstr(string_buffer,"Instr File      Time Received        Time to POCC         Status") == NULL ))
               {
                  fputs(string_buffer, bckqfile);
               }
            }

            /* do not write header to file */
            if (strstr(string_buffer,"Instr File"))
            {
               found_data = TRUE;
            }
         }
         break;         /* out of while */
      }

      /* write event msg (ignore page formats, headers, etc) cmd hist file */
      if ((string_buffer[2] ==  '-' &&
          string_buffer[6] == '-'  &&
          string_buffer[9] == ':'  &&
          string_buffer[12] == ':' &&
          string_buffer[15] == '.') &&
          (!strstr(string_buffer,"CMS Background Queue:")) &&
          (!strstr(string_buffer,"Current Mode Parameter Values")) &&
          (!strstr(string_buffer,"Throughput Mode:") && (!strstr(string_buffer,"RCR Processing:"))) &&
          (!strstr(string_buffer,"Mission:") && (!strstr(string_buffer,"Local GMT:"))) &&
          (!strstr(string_buffer,"Instr File      Time Received        Time to POCC         Status")))
      {
         fputs(string_buffer, cmdfile);
      }
   }  /* end while */

   fclose(infile);
   fclose(cmdfile);
   fclose(bckqfile);

   return(OK);
}  /* end offline_split_cmdhist */



/*********************************************************************
 FUNCTION NAME:         offline_merge_cmdhist
 INPUT  PARAMETERS:     char *filedate, char *scihist2, char *scibck2, 
			char *final_cmd_hist
 OUTPUT PARAMETERS:     char *final_cmd_hist
 RETURN VALUE:          int rc (OK,ERR)
 REFERENCED DATA FILES: final_cmd_hist (merged Cmd Hist SO_CH_NUL_yyyymmdd_Vnn.DAT)
                        scihist1 (cmd hist msgs from SO_CH_NUL_yyyymmdd_Vnn.DAT)
                        scibck1  (bgq msgs from SO_CH_NUL_yyyymmdd_Vnn.DAT)
                        scihist2 (cmd hist msgs from new POCC cmd hist file)
                        scibck2  (bgq msgs from new POCC cmd hist file)
                        scihist3 (cat scihist1 scihist2)
                        scibck3  (cat scibck1  scibck2)
                        scihist4 (sort scihist3)
                        scibck4  (sort scibck3)
                        scihist5 (uniq scihist4)
                        scibck5  (uniq scibck4)
                        final_cmd (scihist5 or 
				   scihist1 if first cmd hist rpt for this day)
                        final_bck (scibck5 or 
				   scibck1  if first cmd hist rpt for this day)

 PURPOSE:               If there is a Cmd Hist file for this day, split 
                        SO_CH_NUL_yyyymmdd_Vnn.DAT into cmd hist and bgq files,
                        merge with new POCC cmd hist and bgq, sort, remove 
			duplicates and create new SO_CH_NUL_yyyymmdd_Vnn.DAT.
                        Otherwise just create new SO_CH_NUL_yyyymmdd_Vnn.DAT.
                                                 
 **************************************************************************/
int offline_merge_cmdhist(char *filedate, char *scihist2, char *scibck2, 
		  char *final_cmd_hist)
{
   int version;
   int rc;
   bool found_data= FALSE;
   FILE *infile;
   FILE *cmdfile;
   FILE *bckqfile;
   FILE *fp;
   char scihist1[MAXBUFFER] = {'\0'};
   char scibck1[MAXBUFFER]  = {'\0'};
   char scihist3[MAXBUFFER] = {'\0'};
   char scibck3[MAXBUFFER]  = {'\0'};
   char scihist4[MAXBUFFER] = {'\0'};
   char scibck4[MAXBUFFER]  = {'\0'};
   char scihist5[MAXBUFFER] = {'\0'};
   char scibck5[MAXBUFFER]  = {'\0'};
   char string_buffer[MAXBUFFER] = {'\0'};
   char final_cmd[MAXBUFFER] = {'\0'};
   char final_bck[MAXBUFFER] = {'\0'};
   char errmsg[MAXBUFFER] = {'\0'};
   char *ptr;
   char buf[MAXBUFFER];

   /* split SO_CH_NUL_yyyymmdd_Vnn.DAT into two files */
   /* find the latest version of this file */
   sprintf(final_cmd_hist,"%s/SO_CH_NUL_%s_V01.DAT",cmd_hist_dir,filedate);

   /* if we have already received a cmd hist file for this day */
   if (infile = fopen(final_cmd_hist,"r"))
   {  /* set up name for all of the temp files */
      sprintf(scihist1,"%s/%s_V01.CHR", temp_files_dir,filedate);
      sprintf(scibck1, "%s/%s_V01.BGQ", temp_files_dir,filedate);
      sprintf(scihist3,"%s/%s_V03.CHR", temp_files_dir,filedate);
      sprintf(scibck3, "%s/%s_V03.BGQ", temp_files_dir,filedate);
      sprintf(scihist4,"%s/%s_V04.CHR", temp_files_dir,filedate);
      sprintf(scibck4, "%s/%s_V04.BGQ", temp_files_dir,filedate);
      sprintf(scihist5,"%s/%s_V05.CHR", temp_files_dir,filedate);
      sprintf(scibck5, "%s/%s_V05.BGQ", temp_files_dir,filedate);

      if ((cmdfile = fopen(scihist1, "w")) == NULL)
      {
         printf("Cannot open %s\n",scihist1);
         return(ERR);
      }

      if ((bckqfile = fopen(scibck1, "w")) == NULL)
      {
         printf("Cannot open %s\n",scibck1);
         return(ERR);
      }

      while (!feof(infile))
      {
         if (fgets(string_buffer,MAXBUFFER-1,infile) == NULL)
         {
            break;
         }
         if (strstr(string_buffer,"COMMAND HISTORY REPORT"))
         {  /* ignore header */
            continue;
         }

         if (strstr(string_buffer, "BACKGROUND QUEUE REPORT"))
         {  /* found end of cmd hist report */
            /* now append bck Q rpt */

            while (!feof(infile))
            {
               if (fgets(string_buffer,MAXBUFFER-1,infile) == NULL)
               {
                  break;
               }
               if (found_data)
               {
                  /* ignore format lines at top\bottom */
                  if (!EQU(string_buffer,"\n") &&
                     (strstr(string_buffer,"END OF BACKGROUND QUEUE REPORT") == NULL) &&
                     (strchr(string_buffer,'*') == NULL))
                  {
                     fputs(string_buffer, bckqfile);
                  }
               }

               /* do not write header to file */
               if (strstr(string_buffer,"Instr File"))
               {
                  found_data = TRUE;
               }
            }
            break;         /* out of while */
         }
         else
         {  /* still in cmd hist portion of the report */
            if (string_buffer[2] ==  '-' &&
                string_buffer[6] == '-'  &&
                string_buffer[9] == ':'  &&
                string_buffer[12] == ':' &&
                string_buffer[15] == '.')
            {  /* ignore format lines at top/bottom */
               fputs(string_buffer, cmdfile);
            }
         }
      }
      fclose(cmdfile);
      fclose(bckqfile);
      fclose(infile);

      /* Append V02 cmd hist to existing V01 files */
      offline_append_files(scihist1, scihist2, scihist3);
#ifndef DEBUG
   /* remove temp files */
   unlink(scihist1);
   unlink(scihist2);

#endif

      /* Append V02 bck que to existing V01 files */
      offline_append_files(scibck1, scibck2, scibck3);
#ifndef DEBUG
   /* remove temp files */
   unlink(scibck1);
   unlink(scibck2);

#endif
      /* Sort file in time-order since POCC may not send files in time-order */
      offline_sort_cmdhist(scihist3, scihist4);
#ifndef DEBUG
   /* remove temp files */
   unlink(scihist3);
#endif

      offline_sort_bgq(scibck3, scibck4);
#ifndef DEBUG
   /* remove temp files */
   unlink(scibck3);
#endif
      /* Get rid of duplicate entries if report start/stop times overlap */
      offline_remove_duplicates(scihist4, scihist5);
#ifndef DEBUG
   /* remove temp files */
   unlink(scihist4);
#endif

      offline_remove_duplicates(scibck4, scibck5);
#ifndef DEBUG
   /* remove temp files */
   unlink(scibck4);
#endif
      /* Create a new SO_CH_NUL_yyyymmdd_Vnn.DAT by */
      /* appending yyyymmdd_V05.CHR and yyyymmdd_V05.BGQ */
      strcpy(final_cmd, scihist5);
      strcpy(final_bck, scibck5);
   } 
   else
   {  /* no need to append, sort, uniq since there was not an existing
         SO_CH_NUL_yyyymmdd_Vnn.DAT file */
      strcpy(final_cmd, scihist2);
      strcpy(final_bck, scibck2);
   }

   offline_format_cmd_hist(final_cmd_hist, final_cmd, final_bck);

#ifndef DEBUG
   /* remove any temp files left */
   unlink(scihist2);
   unlink(scibck2);
   unlink(scihist5);
   unlink(scibck5);
#endif

   return(OK);
}  /* end offline_merge_cmdhist */



/*********************************************************************
 FUNCTION NAME:         offline_get_op_day
 INPUT  PARAMETERS:     char *cmdhist, char *filedate
 OUTPUT PARAMETERS:     char *filedate
 RETURN VALUE:          int OK, ERR
 REFERENCED DATA FILES: None
 PURPOSE:               Convert from POCC file naming convention to EOF. 
 *********************************************************************/
int offline_get_op_day(char *cmdhist, char *filedate)
{
   struct tm *tm_data;
   time_t time_sec;
   int year;
   char jdate[MAXBUFFER];	/* YYDDD format */
   char time_buf[MAXBUFFER];
   char buf[MAXBUFFER];

   time(&time_sec);
   tm_data = gmtime(&time_sec);
   if ((year = tm_data->tm_year) > 90)
   {
      year += 1900;
   }
   else
   {
      year += 2000;
   }
   /* convert from YYDDD to YYMMDD format */
   sprintf(jdate,"%02d%c%c%c", tm_data->tm_year,
                  cmdhist[3], cmdhist[4], cmdhist[5]);

   if (jul2greg(atol(jdate),time_buf) == ERR)
   {  /* filename in wrong format, so use current date/time */
      printf("ERROR POCC/CMS CHR Filename %s - Using Current Date\n",cmdhist);
      sprintf(filedate,"%04d%02d%02d", year,tm_data->tm_mon+1,tm_data->tm_mday);
   }
   else
   {
      sprintf(filedate, "%04d%c%c%c%c",
           year,time_buf[2],time_buf[3],time_buf[4],time_buf[5]);
   }

   return(OK);
}  /* end offline_get_op_day */


/*********************************************************************
 FUNCTION NAME:        offline_format_cmd_hist
 INPUT  PARAMETERS:    char *final_cmd_hist,char *final_cmd,char *final_bck 
 OUTPUT PARAMETERS:    none 
 RETURN VALUE:         int RC, OK 
 REFERENCED DATA FILES: SO_CH_NUL_%s_Vnn.DAT
			yyyymmdd_V05.CHR
			yyyymmdd_V05.BCK
 PURPOSE:               Merge the 'sorted' cmd hist file and 'sorted' 
			bck que file for one operational day into one file. 
 *********************************************************************/
int offline_format_cmd_hist(char *final_cmd_hist, char *final_cmd, char *final_bck)
{
  FILE *fp;     /* SO_CH_NUL_%s_Vnn.DAT file */
  FILE *infile; /* merged cmd hist file(s) */
  char string_buffer[MAXBUFFER];
  char errmsg[MAXBUFFER];

   if (fp = fopen(final_cmd_hist,"w"))
   {
      /* write header */
      fputs("\n          *****         COMMAND HISTORY REPORT         *****\n\n\n",fp);
      /* write command hist rpt */
      if (infile = fopen(final_cmd,"r"))
      {
         while (!feof(infile))
         {
            if (fgets(string_buffer,MAXBUFFER-1,infile) == NULL)
            {
               break;
            }
            fputs(string_buffer,fp);
         }
         fclose(infile);
      }
      else
      {
         printf("ERROR Creating New File %s\n",final_cmd);
      }

      /* write header */
      fputs("\n\n          *****        BACKGROUND QUEUE REPORT      *****\n\n",fp);
      fputs("Instr File      Time Received        Time to POCC         Status\n",fp);

      /* write background Q rpt */
      if (infile = fopen(final_bck,"r"))
      {
         while (!feof(infile))
         {
            if (fgets(string_buffer,MAXBUFFER-1,infile) == NULL)
            {
               break;
            }
            fputs(string_buffer,fp);
         }
         fclose(infile);
      }
      else
      {
         printf("ERROR Creating New File %s\n",final_bck);
      }
      fclose(fp);
   }
   else
   {
      printf("ERROR Creating New File %s\n",final_cmd_hist);
   }
}  /* end offline_format_cmd_hist */


/*********************************************************************
 FUNCTION NAME:         offline_append_files
 INPUT  PARAMETERS:     char *file1, char *file2, char *file3 
 OUTPUT PARAMETERS:     none
 RETURN VALUE:          int OK, ERR
 REFERENCED DATA FILES: file1, file2, file3
 PURPOSE:               Append file2 at the end of file1 to create file3.
 *********************************************************************/
int offline_append_files(char *file1, char *file2, char *file3)
{
   char buf[MAXBUFFER];
   char errmsg[MAXBUFFER];
   int rc;

   sprintf(buf,"cat %s %s > %s", file1, file2, file3);
   rc = system(buf);
   if (WIFEXITED(rc) == 0)
   {
      printf("ERROR %d %s Could not Append %s to %s\n", errno, buf, file2, file1);
      return(ERR);
   }
   return(OK);   
}  /* end offline_append_files */


/*********************************************************************
 FUNCTION NAME:        offline_sort_cmdhist 
 INPUT  PARAMETERS:    char *infile, char *outfile 
 OUTPUT PARAMETERS:    none 
 RETURN VALUE:         int OK, ERR 
 REFERENCED DATA FILES:infile, outfile 
 PURPOSE:              Sort command history report in 
			time-order (earliest to latest). 
 *********************************************************************/
int offline_sort_cmdhist(char *infile, char *outfile)
{
   char buf[MAXBUFFER];
   char general[MAXBUFFER];
   char errmsg[MAXBUFFER];
   int rc;

   /* Sort the file in time-order */
   sprintf(buf,"sort -n -t- -t:  %s > %s", infile, outfile);
   rc = system(buf);
   if (WIFEXITED(rc) == 0)
   {
      printf("ERROR %d %s Could Not Sort %s\n", errno, buf, infile);
      return(ERR);
   }
   return(OK);
}  /* end offline_sort_cmdhist */


/*********************************************************************
 FUNCTION NAME:        offline_sort_bgq 
 INPUT  PARAMETERS:    char *infile, char *outfile 
 OUTPUT PARAMETERS:    none 
 RETURN VALUE:         int RC, OK 
 REFERENCED DATA FILES:infile, outfile 
 PURPOSE:              Sort background queue report in time
			order (earliest to latest). 
 *********************************************************************/
int offline_sort_bgq(char *infile, char *outfile)
{
   char buf[MAXBUFFER];
   char errmsg[MAXBUFFER];
   int rc;

   sprintf(buf,"sort -n -k 2.1b,2 %s > %s", infile, outfile);
   rc = system(buf);
   if (WIFEXITED(rc) == 0)
   {
      printf("ERROR %d %s Could Not Sort %s\n", errno, buf, infile);
      return(ERR);
   }
   return(OK);
}  /* end offline_sort_bgq */


/*********************************************************************
 FUNCTION NAME:         offline_remove_duplicates
 INPUT  PARAMETERS:     char *infile, char *outfile 
 OUTPUT PARAMETERS:     none  
 RETURN VALUE:          int OK, ERR
 REFERENCED DATA FILES: infile, outfile
 PURPOSE:               Remove duplicate entries in file.  Note the
			file must be sorted before calling this function.
 *********************************************************************/
int offline_remove_duplicates(char *infile, char *outfile)
{
   char buf[MAXBUFFER];
   char errmsg[MAXBUFFER];
   int rc;

   sprintf(buf,"uniq %s %s", infile, outfile);
   rc = system(buf);
   if (WIFEXITED(rc) == 0)
   {
      printf("ERROR %d %s Could Not Remove Duplicates from %s\n", errno, buf, infile);
      return(ERR);
   }
   return(OK);
}  /* end offline_remove_duplicates */
